@@ -1,94 +1,82 @@
-#' computes bootstrap resamples of your data,
+#' Computes bootstrap resamples of your data,
 #' stores estimates + SEs.
 #'
 #' By default, this will compute bootstrap resamples
-#' and then send them to `bootstrap_ci`
-#' for calculation. Note - only use parallel methods if your
-#' model is expensive to build, otherwise the overhead won't be worth it.
-#'
+#' and then send them to \code{bootstrap_ci}
+#' for calculation.
 #' @import methods
 #' @import stats
-#'
-#' @param base_model
-#'   The pre-bootstrap model, i.e. the model output
+#' @param base_model The pre-bootstrap model, i.e. the model output
 #'   from running a standard model call.
 #'   Examples:
-#'   base_model <- glmmTMB(y ~ age + (1 | subj),
-#'                         data = rel_data, family = binomial)
-#'   base_model <- lm(y ~ x, data = xy_frame)
-#'
-#' @param base_data
-#'   The data that was used in the call. You
+#'   \code{base_model <- glmmTMB(y ~ age + (1 | subj),
+#'                         data = rel_data, family = binomial)}
+#'   \code{base_model <- lm(y ~ x, data = xy_frame)}
+#' @param base_data The data that was used in the call. You
 #'   can leave this to be automatically read, but
 #'   I highly recommend supplying it
-#'
-#' @param resamples
-#'   How many resamples of your data do you want to do?
-#'   9999 is a reasonable default (see Hesterberg 2015),
+#' @param resamples How many resamples of your data do you want to do?
+#'   9,999 is a reasonable default (see Hesterberg 2015),
 #'   but start very small to make sure it works on
 #'   your data properly, and to get a rough timing estimate etc.
-#'
-#' @param return_coefs_instead
-#'   Logical, default FALSE: do you want the list of lists
-#'   of results for each bootstrap sample (set to TRUE), or the
-#'   matrix output of all samples? See return for more details.
-#'
-#' @param parallelism
-#'   What type of parallelism (if any) to use to run the resamples.
+#' @param return_coefs_instead Logical, default \code{FALSE}: do you want the
+#'   list of lists of results for each bootstrap sample (set to \code{TRUE}),
+#'   or the matrix output of all samples? See return for more details.
+#' @param parallelism Type of parallelism (if any) to use to run the resamples.
 #'   Options are:
-#'   - "none"      the default
-#'   - "future"    to use future.apply (`future`s)
-#'   - "parallel"  to use parallel::mclapply
-#'
-#' @param resample_specific_blocks
-#'   Character vector, default NULL. If left NULL,
-#'   this algorithm with choose ONE random block to resample over -
+#'   \describe{
+#'     \item{\code{"none"}}{The default, sequential}
+#'     \item{\code{"future"}}{To use \code{future.apply} (\code{future}s)}
+#'     \item{\code{"parallel"}}{To use \code{parallel::mclapply}}
+#'   }
+#' @param resample_specific_blocks Character vector, default \code{NULL}.
+#'   If left \code{NULL}, this algorithm with choose ONE random block to resample over -
 #'   the one with the largest entropy (often the one with most levels).
-#'   If you wish to
-#'   resample over specific random effects as blocks, enter
+#'   If you wish to resample over specific random effects as blocks, enter
 #'   the names here - can be one, or many. Note that resampling
 #'   multiple blocks is in general quite conservative.
-#'
 #'   If you want to perform case resampling but you DO have
-#'   random effects, set resample_specific_blocks to any
+#'   random effects, set \code{resample_specific_blocks} to any
 #'   non-null value that isn't equal to a random effect
 #'   variable name.
-#'
-#' @param unique_resample_lim
-#'   Should be same length as number of random effects (or left NULL).
+#' @param unique_resample_lim Should be same length as number of random effects
+#'   (or left \code{NULL}).
 #'   Do you want to force the resampling to produce a minimum number of
-#'   unique values in sampling? Don't make this too big...
+#'   unique values in sampling? Don't make this too big.
 #'   Must be named same as rand cols
-#'
-#' @param narrowness_avoid
-#'   Boolean, default TRUE.
-#'   If TRUE, will resample n-1 instead of n elements
-#'   in the bootstrap (n being either rows, or random effect levels,
-#'   depending on existence of random effects). If FALSE, will do
+#' @param narrowness_avoid Boolean, default \code{TRUE}. If \code{TRUE}, will resample n-1
+#'   instead of n elements in the bootstrap (n being either rows,
+#'   or random effect levels,
+#'   depending on existence of random effects). If \code{FALSE}, will do
 #'   typical size n resampling.
-#'
-#' @param num_cores
-#'   Defaults to parallel::detectCores() - 1 if parallelism = "parallel"
-#'
-#' @param suppress_sampling_message
-#'   Logical, default FALSE. By default, this function
-#'   will message the console with the type of bootstrapping:
-#'   block resampling over random effects - in which case it'll say
-#'   what effect it's sampling over;
-#'   case resampling - in which case it'll say as much.
-#'   Set TRUE to hide message.
-#'
-#' @return
-#'   By default, returns the output from bootstrap_ci:
-#'     - for each set of covariates (usually just the one set,
-#'       the conditional model), a matrix of output, a row for each variable,
-#'       including the intercept
-#'       (estimate, CIs for boot and base, p-values).
-#'   If return_coefs_instead = TRUE, then will instead
+#' @param num_cores How many cores to use.
+#'   Defaults to \code{parallel::detectCores() - 1L} if
+#'   \code{parallelism = "parallel"}
+#' @param future_packages Packages to pass to created futures when
+#'   using \code{parallelism = "future"}. This must be supplied if
+#'   the package used to model the data isn't in base and you're
+#'   using a plan that doesn't have shared memory, because the
+#'   model is updated with the S3 generic \code{update}.
+#' @param suppress_sampling_message Logical, the default is
+#'   to suppress if not in an interactive session.
+#'   Do you want the function to message the console with the type of
+#'   bootstrapping? If block resampling over random effects, then it'll say
+#'   what effect it's sampling over; if case resampling -
+#'   in which case it'll say as much.
+#'   Set \code{TRUE} to hide message.
+#' @return By default (with \code{return_coefs_instead} being \code{FALSE}),
+#'   returns the output from \code{bootstrap_ci};
+#'   for each set of covariates (usually just the one set,
+#'   the conditional model) we get a matrix of output: a row for each variable
+#'   (including the intercept),
+#'   estimate, CIs for boot and base, p-values.
+#'   If \code{return_coefs_instead} is \code{TRUE}, then will instead
 #'   return a list of length two:
-#'   [[1]] will be a list containing the output for the base model
-#'   [[2]] will be a list of length resamples,
-#'   each a list of matrices of estimates and standard errors for each model.
+#'   \itemize{
+#'     \item a list containing the output for the base model
+#'     \item a list of length \code{resamples} each a list of matrices of
+#'       estimates and standard errors for each model.
+#'   }
 #'   This output is useful for error checking, and if you want
 #'   to run this function in certain distributed ways.
 #'
@@ -127,12 +115,9 @@ bootstrap_model <- function(base_model,
                             unique_resample_lim = NULL,
                             narrowness_avoid = TRUE,
                             num_cores = NULL,
-                            suppress_sampling_message = FALSE){
+                            future_packages = NULL,
+                            suppress_sampling_message = !interactive()){
     if (missing(base_data) || is.null(base_data)) {
-        warning("Please supply data through the argument base_data; ",
-                "automatic reading from your model can produce ",
-                "unforeseeable bugs.", call. = FALSE)
-
         if ("model" %in% names(base_model)) {
             base_data <- base_model$model
         } else if ("frame" %in% names(base_model)) {
@@ -144,6 +129,10 @@ bootstrap_model <- function(base_model,
                  "please supply data as base_data ",
                  "to this function", call. = FALSE)
         }
+
+        warning("Please supply data through the argument base_data; ",
+                "automatic reading from your model can produce ",
+                "unforeseeable bugs.", call. = FALSE)
     }
 
     if (missing(parallelism)) {
@@ -183,9 +172,20 @@ bootstrap_model <- function(base_model,
                 stop("`parallelism = \"parallel\"` uses `package:parallel`, ",
                      "but it's not installed", call. = FALSE)
             } # nocov end
+
+            if (is.null(num_cores)) { # nocov start
+                num_cores <- max(parallel::detectCores() - 1L, 1L)
+                message("`num_cores` not set, defaulting to ", num_cores,
+                        " (`parallel::detectCores() - 1L`)")
+            } # nocov end
         }
     }
 
+    if (parallelism != "future" && !is.null(future_packages)) {
+        stop("Argument `future_packages` should only be set when ",
+             "using `parallelism = \"future\"`", call. = FALSE)
+    }
+
     ##------------------------------------
 
     ## formula processing
@@ -307,9 +307,10 @@ bootstrap_model <- function(base_model,
     ##------------------------------------
 
     coef_se_list <- bootstrap_runner(bootstrap_coef_est,
-                                     resamples,
-                                     parallelism,
-                                     num_cores)
+                                     resamples = resamples,
+                                     parallelism = parallelism,
+                                     num_cores = num_cores,
+                                     future_packages = future_packages)
 
     ##------------------------------------
 
@@ -322,7 +323,7 @@ bootstrap_model <- function(base_model,
     }
 
     ## keep going until solved
-    max_redos <- 10L
+    max_redos <- 15L
     redo_iter <- 1L
     while (sum(error_ind) > 0L && redo_iter <= max_redos) {
         message(sum(error_ind), " error(s) to redo")
@@ -390,33 +391,23 @@ BootGlmm <- function(base_model, # nocov start
                     suppress_sampling_message = suppress_sampling_message)
 } # nocov end
 
-#' Runs the bootstrapping of the models
+
+#' Runs the bootstrapping of the models.
 #'
 #' This function gets passed a function that runs a single bootstrap resample
 #' and a number of resamples, and decides how to run them
-#' e.g. in parallel etc
-#'
-#' @param bootstrap_function
-#'   a function that we wish to run `resamples` times
-#'
-#' @param resamples
-#'   how many times we wish to run `bootstrap_function`
-#'
-#' @param parallelism
-#'   How to run this function in parallel
-#'
-#' @param num_cores
-#'   How many cores to use, if using `parallel`
-#'
-#' @return
-#'   returns the list that contains the results of running
-#'   `bootstrap_function`.
-#'
+#' e.g. in parallel etc.
+#' @param bootstrap_function Function that we wish to run
+#'   \code{resamples} times.
+#' @inheritParams bootstrap_model
+#' @return Returns the list that contains the results of running
+#'   \code{bootstrap_function}.
 #' @keywords internal
 bootstrap_runner <- function(bootstrap_function,
                              resamples,
                              parallelism = c("none", "future", "parallel"),
-                             num_cores = NULL){
+                             num_cores = NULL,
+                             future_packages = NULL){
     parallelism <- match.arg(parallelism)
 
     if (parallelism == "future") {
@@ -425,7 +416,8 @@ bootstrap_runner <- function(bootstrap_function,
                                  1:resamples,
                                  function(i){
                                      bootstrap_function()
-                                 }))
+                                 },
+                                 future.packages = future_packages))
     }
 
     if (parallelism == "parallel") {
